% MAIN_SCRIPT.M: This is the main script for running the analysis
% presented in Beard and Qian, Thermodynamic-based computational
% profiling of cellular regulatory control in hepatocyte metabolism.
% Am J. Physiol. Endocrinol. Metab. 288:E633-E644, 2005.
%
% To run the script, input the commands below to the MATLAB
% command prompt, which call a series of functions that are
% included in this package.
%
% Copyright Daniel A. Beard
% Medical College of Wisconsin
% (414) 456-5752
% dbeard@mcw.edu
%
% It is requested that any use of these codes, or adapations of these 
% codes, references the above publication.
%

% STEP 1.
% Loading network: using the function "read_bcn_file" to input the network
%   data from the file "glycogen_met.bcn". The output is the following 
%   variables:
%   (1) species: (string array) list of species in the network
%   (2) enzymes: (string array) list of enzyme names
%   (3) genes: (string array) list of gene abbreviations
%   (4) reversibility: (boolean) list of rxn. reversibility
%   (5) S: (real array) stoichiometric matrix
%   (6) EC: (string array) EC numbers
[species,enzymes,genes,reversibility,S,EC] = ...
    read_bcn_file('glycogen_met.w.bcn');
[m,n] = size(S);
n_int = n; % number of internal reactions

% STEP 2.
% Computing the cycles in the network:
C = cycles(S);

% Adding columns for transport fluxes to the stoichiometric matrix:
% Input fluxes
S(:,n+1) = zeros(m,1); S(1 ,n+1) = 1;   % glucose input
S(:,n+2) = zeros(m,1); S(48,n+2) = 1;   % oxygen input
% Output fluxes:
S(:,n+3) = zeros(m,1); S(28,n+3) = -1;  % CO2 output
S(:,n+4) = zeros(m,1); S(22,n+4) = -1;  % glycogen accumulation
% Water flux:
S(:,n+5) = zeros(m,1); S(16,n+5) = 1;   % water flux
% Indexes for transport fluxes:
j_GLC   = n+1;
j_O2    = n+2;
j_CO2   = n+3;
j_glyco = n+4;
j_water = n+5;
j_ATPM  = 41;    % ATP use
[m,n] = size(S);
Srx = S(:,1:n_int); % S matrix for internal reactions

% STEP 3.
% Assigning partition coefficients (these are ballpark guesses, 
% but the results are not highly sensitive to the exact values.)
d1 = 0.75; % ml water space (ml tissue)^{-1}
d2 = 0.25; % ml water space (ml tissue)^{-1}
d = ones(m,1);
d([1:23 28 48:51 54:55]) = d1;  % cytosol space
d([24:27 29:47 52:53]) = d2;    % mitochondrial space

% STEP 4.
% Computing conserved concentration pools:
L = null((diag(1./d)*S)','r');

% STEP 5.
% Assigning standard potentials (transformed free energies of
% formation (ref. Alberty's book for most of these values.)
go(1) = -426.71;    %GLC
go(2) = -2292.50;   %ATP
go(3) = -1318.92;   %G6P
go(4) = -1424.70;   %ADP
go(5) = -1315.74;   %F6P
go(6) = -2206.78;   %FDP
go(7) = -1088.04;   %T3P1
go(8) = -1095.70;   %T3P2
go(9) = -1059.49;   %PI
go(10) = 1059.11;   %NAD
go(11) = 1120.09;   %NADH
go(12) = -2207.30;  %13DPG
go(13) = -1347.73;  %3PG
go(14) = -1341.79;  %2PG
go(15) = -1189.73;  %PEP
go(16) = -155.66;   %H2O
go(17) = -350.78;   %PYR
go(18) = -1311.89;  %G1P
go(19) = -2292.50;  %UTP (taken as ATP potential)
go(20) = -1647.94;  %UDPG (taken to fix dG_o(udp1) = 3.77 kcal mol^{-1}
go(21) = -1940.66;  %PPI
go(22) = -248.02;   %GLYCOGEN ( glycogen[n] -> glycogen[n+1] )
go(23) = -1424.70;  %UDP (taken as ADP potential)
go(24) = -350.78;   %PYR_m
go(25) = -7.26;     %COA_m
go(26) = 1059.11;   %NAD_m
go(27) = 1120.09;   %NADH_m
go(28) = -547.10;   %CO2
go(29) = -66.22;    %ACCOA_m
go(30) = -715.00;   %OA_m
go(31) = -966.23;   %CIT_m
go(32) = -959.58;   %ICIT_m
go(33) = -633.59;   %AKG_m
go(34) = -347.34;   %SUCCOA_m
go(35) = -1424.70;  %GDP_m (taken as ADP potential)
go(36) = -1059.49;  %PI_m
go(37) = -2292.50;  %GTP_m (taken as ATP potential)
go(38) = -530.64;   %SUCC_m
go(39) = 1260.51;   %FAD_m
go(40) = 1253.44;   %FADH_m
go(41) = -523.58;   %FUM_m
go(42) = -682.85;   %MAL_m
go(43) = -1424.70;  %ADP_m
go(44) = -2292.50;  %ATP_m
go(45) = -7.29;     %Q
go(46) = -15.44;    %H_m
go(47) = -27.75;    %QH2
go(48) = 16.40;     %O2
go(49) = -715.00;   %OA
go(50) = -682.85;   %MAL
go(51) = -633.59;   %AKG
go(52) = -372.15;   %GLU_m
go(53) = -452.09;   %ASP_m
go(54) = -372.15;   %GLU
go(55) = -452.09;   %ASP
go = go./4.1871;  % Convert to kcal mol^{-1}
kT = 0.5950;
dGo = Srx'*go';

% List the computed \Delta G's for the reactions:
[(1:n_int)' (Srx'*go').*4.1871]

% STEP 6.
% Computing the fluxes for the two metabolic modes.
% Nonlinear optimizer settings:
opts = optimset('fmincon');
opts = optimset(opts,...
    'TolCon',1e-9,...
    'TolFun',1e-9,...
    'TolPCG',1e-3,...
    'TolX',1e-9,...
    'Display','iter',...
    'Maxiter',1000,...
    'MaxFunEvals',10000);

% 6.1. Computing Mode 1: Glycogen Production from Glucose
LB = -Inf.*ones(n,1);
UB = +Inf.*ones(n,1);
for j = 1:length(reversibility)
  if ~reversibility(j)
    LB(j) = 0;    
  end
end
% Bounds on transport fluxes:
LB(n_int+3) = 0; % constrains CO2 flux to be output
UB(j_GLC) = 1.5; % from experiment
LB(j_ATPM) = 1.44; % from experiment
j_obj = j_glyco; % this flux will be maximized
Jo = ones(n,1);  % initial guess
J1 = fmincon(@max_flux,Jo,[],[],S,zeros(m,1),LB,UB,...
    @T_constraint,[],C,j_obj);

% Since the nonlinear optimizer is flaky, we use the final solution
% as the initial guess and re-run the optimization using the following
% three lines of code. Sometimes you may have to run this a few times
% until the solution converges.
  Jo = J1;
  J1 = fmincon(@max_flux,Jo,[],[],S,zeros(m,1),LB,UB,...
      @T_constraint,[],C,j_obj);
  J1 % list J_1 the flux vector for Mode 1

% 6.2. Computing Mode 2: Glucose Production from Glycogen
LB = -Inf.*ones(n,1);
UB = +Inf.*ones(n,1);
for j = 1:length(reversibility)
  if ~reversibility(j)
    LB(j) = 0;    
  end
end
% Bounds on transport fluxes:
LB(n_int+3) = 0; % constrains CO2 flux to be output
LB(j_glyco) = -3.8340;  % (This is the maximal rate of glycogenolysis
                        % that will give the correct glucose output
                        % from experiment.)
UB(j_O2) = 2.45;    % from experiment
LB(j_ATPM) = 1.44;  % from experiment
j_obj = j_GLC; % this flux will be minimized
J2 = fmincon(@min_flux,J1,[],[],S,zeros(m,1),LB,UB,...
    @T_constraint,[],C,j_obj);

% As above, may have to re-run optimizer serveral times to
% get convergence.
  Jo = J2;
  J2 = fmincon(@min_flux,Jo,[],[],S,zeros(m,1),LB,UB,...
      @T_constraint,[],C,j_obj);
  J2 % list J_2 the flux vector for Mode 1
  
% Assigning directions of flux vectors:
for i = 1:n_int
  if abs(J1(i)) < 1e-6
      J1(i) = 0;
  end
  if abs(J2(i)) < 1e-6
      J2(i) = 0;
  end
end
dir1 = sign(J1(1:n_int)); % directions for Mode 1
dir2 = sign(J2(1:n_int)); % directions for Mode 2

% Step 7. Calculations of concentrations and activities
warning off MATLAB:divideByZero

% Constructing constraint matrix and vector for
% constraint on Phosphate potential (constrainted to be
% between -13 and -15 kcal per mole). This constraint
% is applied A*g < b, where "g" is the vector of potentials.
A = [Srx(:,j_ATPM)' zeros(1,m)
     zeros(1,m) Srx(:,j_ATPM)'
     -Srx(:,j_ATPM)' zeros(1,m)
     zeros(1,m) -Srx(:,j_ATPM)'];
b = [-13 -13 15 15]';

% 7.1 Minimizing differences between concentrations.
% (These optimizations can take some time. There may 
% be room for improvement in the efficiency.)
% Nonlinear optimizer settings:
opts = optimset('fmincon');
opts = optimset(opts,...
    'TolCon',1e-4,...
    'TolFun',1e-6,...
    'TolPCG',1e-3,...
    'TolX',1e-4,...
    'Display','iter',...
    'Maxiter',1e3,...
    'MaxSQPIter',1e3,...
    'MaxFunEvals',1e5);
[g,F] = fmincon(@opt_fun1,[go go]'+kT*log(1e-3),A,b,[],[],[],[],...
    @gfe_cons,opts,J1,J2,Srx,dir1,dir2,go,kT,L);
g_ob1 = g; % predicted vector of potentials using objective 1.

% 7.2a Minimizing differences between activities.
% (This calculation can take Run optimizer several times to get lowest possible minima.)
opts = optimset('fmincon');
opts = optimset(opts,...
    'TolCon',1e-4,...
    'TolFun',1e-2,...
    'TolPCG',1e-3,...
    'TolX',1e-4,...
    'Display','iter',...
    'Maxiter',1e3,...
    'MaxSQPIter',1e3,...
    'MaxFunEvals',1e5);  
[g,F] = fmincon(@opt_fun2,[go go]'+kT*log(1e-3),A,b,[],[],[],[],...
    @gfe_cons,opts,J1,J2,Srx,dir1,dir2,go,kT,L);
g_ob2 = g;

% Step 8. Plotting the results.

% 8.1 Plotting results for steady state 1:
g = g_ob1; plot_script;

% 8.2 Plotting results for steady state 2:
g = g_ob2; plot_script;


